/*
 * [MASMState.java]
 *
 * Summary: State machine for parsing MASM to align the columns.
 *
 * Copyright: (c) 2004-2017 Roedy Green, Canadian Mind Products, http://mindprod.com
 *
 * Licence: This software may be copied and used freely for any purpose but military.
 *          http://mindprod.com/contact/nonmil.html
 *
 * Requires: JDK 1.8+
 *
 * Created with: JetBrains IntelliJ IDEA IDE http://www.jetbrains.com/idea/
 *
 * Version History:
 *  1.0 2011-01-01 initial version.
 */
package com.mindprod.masmbalancer;

import com.mindprod.common18.ST;
import com.mindprod.entities.EntifyStrings;
import com.mindprod.fastcat.FastCat;

import java.util.regex.Pattern;

import static java.lang.System.*;

/**
 * State machine for parsing MASM to align the columns.
 *
 * @author Roedy Green, Canadian Mind Products
 * @version 1.0 2011-01-02 initial version
 * @since 2011-01-02
 */
public enum MASMState
    {
        AT_FIRST_COLUMN
                    {
                    void consume( char c )
                        {
                        // ignore spaces.
                        }

                    void leaving()
                        {
                        // just ignore spaces.
                        }

                    MASMState next( MASMCharCategory category,
                                    char nextChar,
                                    boolean first )
                        {
                        switch ( category )
                            {
                            case EOL:
                                how = HowToProcess.FORWARD;
                                return AT_END_OF_LINE;
                            case SPACE:
                                how = HowToProcess.DISCARD;
                                return SEEKING_OPERATOR;
                            case PLAIN:
                                if ( isComing( "comment " ) )  // case-insensitive
                                    {
                                    how = HowToProcess.FORWARD;
                                    return IN_MULTILINE_COMMENT;
                                    }
                                else
                                    {
                                    how = HowToProcess.FORWARD;
                                    return IN_LABEL;
                                    }
                            case LT:
                            case GT:
                            case QUOTE:
                            case TICK:
                                how = HowToProcess.FORWARD;
                                return IN_LABEL;
                            case SEMICOLON:
                                how = HowToProcess.FORWARD;
                                return IN_LINE_COMMENT;
                            default:
                                assert false :
                                        "bad state "
                                        + category
                                        + " "
                                        + nextChar;
                                return null;
                            } // end switch
                        }
                    },
        IN_LABEL
                    {
                    @SuppressWarnings( { "UnusedParameters" } )
                    void consume( char c )
                        {
                        accumulatedLabel.append( c );
                        }

                    void leaving()
                        {
                        // just leave Accumulated label till end of line..
                        }

                    MASMState next( MASMCharCategory category,
                                    char nextChar,
                                    boolean first )
                        {
                        switch ( category )
                            {
                            case EOL:
                                how = HowToProcess.FORWARD;
                                return AT_END_OF_LINE;
                            case SPACE:
                                how = HowToProcess.DISCARD;
                                return SEEKING_OPERATOR;
                            case PLAIN:
                            case QUOTE:
                            case TICK:
                                how = HowToProcess.CONSUME;
                                return IN_LABEL;
                            case SEMICOLON:
                                how = HowToProcess.FORWARD;
                                return IN_TAIL_COMMENT;
                            default:
                                assert false :
                                        "bad state "
                                        + category
                                        + " "
                                        + nextChar;
                                return null;
                            } // end switch}
                        }
                    },
        SEEKING_OPERATOR
                    {
                    @SuppressWarnings( { "UnusedParameters" } )
                    void consume( char c )
                        {
                        // just ignore the spaces.
                        }

                    void leaving()
                        {
                        // just ignore spaces.
                        }

                    MASMState next( MASMCharCategory category,
                                    char nextChar,
                                    boolean first )
                        {
                        switch ( category )
                            {
                            case EOL:
                                how = HowToProcess.FORWARD;
                                return AT_END_OF_LINE;
                            case SPACE:
                                how = HowToProcess.DISCARD;
                                return SEEKING_OPERATOR;
                            case LT:
                            case GT:
                            case PLAIN:
                            case QUOTE:
                            case TICK:
                                how = HowToProcess.FORWARD;
                                return IN_OPERATOR;
                            case SEMICOLON:
                                how = HowToProcess.FORWARD;
                                return IN_TAIL_COMMENT;
                            default:
                                assert false :
                                        "bad state "
                                        + category
                                        + " "
                                        + nextChar;
                                return null;
                            } // end switch
                        }
                    },
        IN_OPERATOR
                    {
                    void consume( char c )
                        {
                        accumulatedOperator.append( c );
                        }

                    void leaving()
                        {
                        // just leave Accumulated operator till end of line..
                        }

                    MASMState next( MASMCharCategory category,
                                    char nextChar,
                                    boolean first )
                        {
                        switch ( category )
                            {
                            case EOL:
                                how = HowToProcess.FORWARD;
                                return AT_END_OF_LINE;
                            case SPACE:
                                how = HowToProcess.DISCARD;
                                return SEEKING_OPERAND;
                            case LT:
                            case GT:
                            case PLAIN:
                            case QUOTE:
                            case TICK:
                                how = HowToProcess.CONSUME;
                                return IN_OPERATOR;
                            case SEMICOLON:
                                how = HowToProcess.FORWARD;
                                return IN_TAIL_COMMENT;
                            default:
                                assert false :
                                        "bad state "
                                        + category
                                        + " "
                                        + nextChar;
                                return null;
                            } // end switch}
                        }
                    },
        SEEKING_OPERAND
                    {
                    @SuppressWarnings( { "UnusedParameters" } )
                    void consume( char c )
                        {
                        // just ignore the spaces.
                        }

                    void leaving()
                        {
                        // just ignore spaces.
                        }

                    @SuppressWarnings( { "UnusedDeclaration" } )
                    MASMState next( MASMCharCategory category,
                                    char nextChar,
                                    boolean first )
                        {
                        switch ( category )
                            {
                            case EOL:
                                how = HowToProcess.FORWARD;
                                return AT_END_OF_LINE;
                            case SPACE:
                                how = HowToProcess.DISCARD;
                                return SEEKING_OPERAND;
                            case LT:
                            case GT:
                            case PLAIN:
                            case QUOTE:
                            case TICK:
                                how = HowToProcess.FORWARD;
                                return IN_OPERAND;
                            case SEMICOLON:
                                how = HowToProcess.FORWARD;
                                return IN_TAIL_COMMENT;
                            default:
                                assert false :
                                        "bad state "
                                        + category
                                        + " "
                                        + nextChar;
                                return null;
                            } // end switch
                        }
                    },
        IN_OPERAND
                    {
                    void consume( char c )
                        {
                        accumulatedOperand.append( c );
                        }

                    void leaving()
                        {
                        // just leave Accumulated operator till end of line..
                        }

                    MASMState next( MASMCharCategory category,
                                    char nextChar,
                                    boolean first )
                        {
                        switch ( category )
                            {
                            case EOL:
                                how = HowToProcess.FORWARD;
                                return AT_END_OF_LINE;
                            case GT:
                            case PLAIN:
                            case SPACE:
                                how = HowToProcess.CONSUME;
                                return IN_OPERAND;
                            case LT:
                                how = HowToProcess.FORWARD;
                                return IN_ANGLE_OPERAND;
                            case QUOTE:
                                how = HowToProcess.FORWARD;
                                return IN_QUOTE_OPERAND;
                            case TICK:
                                how = HowToProcess.FORWARD;
                                return IN_TICK_OPERAND;
                            case SEMICOLON:
                                how = HowToProcess.FORWARD;
                                return IN_TAIL_COMMENT;
                            default:
                                assert false :
                                        "bad state "
                                        + category
                                        + " "
                                        + nextChar;
                                return null;
                            } // end switch}
                        }
                    },
        IN_ANGLE_OPERAND
                    {
                    void consume( char c )
                        {
                        accumulatedOperand.append( c );
                        }

                    void leaving()
                        {
                        // just leave Accumulated operator till end of line..
                        }

                    MASMState next( MASMCharCategory category,
                                    char nextChar,
                                    boolean first )
                        {
                        switch ( category )
                            {
                            case EOL:
                                how = HowToProcess.FORWARD;
                                return AT_END_OF_LINE;
                            case LT:
                            case PLAIN:
                            case QUOTE:
                            case SEMICOLON:
                            case SPACE:
                            case TICK:
                                how = HowToProcess.CONSUME;
                                return IN_ANGLE_OPERAND;
                            case GT:
                                how = HowToProcess.CONSUME;
                                return IN_OPERAND;
                            default:
                                assert false :
                                        "bad state "
                                        + category
                                        + " "
                                        + nextChar;
                                return null;
                            } // end switch}
                        }
                    },
        IN_TICK_OPERAND
                    {
                    void consume( char c )
                        {
                        accumulatedOperand.append( c );
                        }

                    void leaving()
                        {
                        // just leave Accumulated operator till end of line..
                        }

                    MASMState next( MASMCharCategory category,
                                    char nextChar,
                                    boolean first )
                        {
                        switch ( category )
                            {
                            case EOL:
                                how = HowToProcess.FORWARD;
                                return AT_END_OF_LINE;
                            case LT:
                            case GT:
                            case PLAIN:
                            case QUOTE:
                            case SEMICOLON:
                            case SPACE:
                                how = HowToProcess.CONSUME;
                                return IN_TICK_OPERAND;
                            case TICK:
                                how = HowToProcess.CONSUME;
                                return IN_OPERAND;
                            default:
                                assert false :
                                        "bad state "
                                        + category
                                        + " "
                                        + nextChar;
                                return null;
                            } // end switch}
                        }
                    },
        IN_QUOTE_OPERAND
                    {
                    void consume( char c )
                        {
                        accumulatedOperand.append( c );
                        }

                    void leaving()
                        {
                        // just leave Accumulated operator till end of line..
                        }

                    MASMState next( MASMCharCategory category,
                                    char nextChar,
                                    boolean first )
                        {
                        switch ( category )
                            {
                            case EOL:
                                how = HowToProcess.FORWARD;
                                return AT_END_OF_LINE;
                            case LT:
                            case GT:
                            case PLAIN:
                            case SEMICOLON:
                            case SPACE:
                            case TICK:
                                how = HowToProcess.CONSUME;
                                return IN_QUOTE_OPERAND;
                            case QUOTE:
                                how = HowToProcess.CONSUME;
                                return IN_OPERAND;
                            default:
                                assert false :
                                        "bad state "
                                        + category
                                        + " "
                                        + nextChar;
                                return null;
                            } // end switch}
                        }
                    },
        IN_LINE_COMMENT
                    {
                    void consume( char c )
                        {
                        accumulatedLineComment.append( c );
                        }

                    void leaving()
                        {
                        // just leave Accumulated tail comment till end of line..
                        }

                    MASMState next( MASMCharCategory category,
                                    char nextChar,
                                    boolean first )
                        {
                        switch ( category )
                            {
                            case EOL:
                                how = HowToProcess.FORWARD;
                                return AT_END_OF_LINE;
                            case LT:
                            case GT:
                            case PLAIN:
                            case QUOTE:
                            case SEMICOLON:
                            case SPACE:
                            case TICK:
                                how = HowToProcess.CONSUME;
                                return IN_LINE_COMMENT;
                            default:
                                assert false :
                                        "bad state "
                                        + category
                                        + " "
                                        + nextChar;
                                return null;
                            } // end switch}
                        }
                    },
        IN_TAIL_COMMENT
                    {
                    void consume( char c )
                        {
                        accumulatedTailComment.append( c );
                        }

                    void leaving()
                        {
                        // just leave Accumulated tail comment till end of line..
                        }

                    MASMState next( MASMCharCategory category,
                                    char nextChar,
                                    boolean first )
                        {
                        switch ( category )
                            {
                            case EOL:
                                how = HowToProcess.FORWARD;
                                return AT_END_OF_LINE;
                            case LT:
                            case GT:
                            case PLAIN:
                            case QUOTE:
                            case SEMICOLON:
                            case SPACE:
                            case TICK:
                                how = HowToProcess.CONSUME;
                                return IN_TAIL_COMMENT;
                            default:
                                assert false :
                                        "bad state "
                                        + category
                                        + " "
                                        + nextChar;
                                return null;
                            } // end switch}
                        }
                    },
        IN_MULTILINE_COMMENT
                    {
                    char delimiter;

                    void consume( char c )
                        {
                        accumulatedMultilineComment.append( c );
                        }

                    void leaving()
                        {
                        // just leave Accumulated tail comment till end of line..
                        }

                    MASMState next( MASMCharCategory category,
                                    char nextChar,
                                    boolean first )
                        {
                        if ( first )
                            {
                            delimiter = 0;
                            }
                        switch ( category )
                            {
                            case EOL:
                                how = HowToProcess.CONSUME;
                                return IN_MULTILINE_COMMENT;
                            case PLAIN:
                                if ( delimiter == 0 && !ST.isLegal( nextChar, "coment COMENT" ) )   // do NOT correct to
                                    // COMMENT
                                    {
                                    delimiter = nextChar;  // first non-blank after common
                                    }
                                else if ( nextChar == delimiter )
                                    {
                                    // found end of comment
                                    how = HowToProcess.CONSUME;
                                    return AT_FIRST_COLUMN;  // it is not really the first column, but logically it is.
                                    }
                                how = HowToProcess.CONSUME;
                                return IN_MULTILINE_COMMENT;
                            case LT:
                            case GT:
                            case QUOTE:
                            case SEMICOLON:
                            case SPACE:
                            case TICK:
                                how = HowToProcess.CONSUME;
                                return IN_MULTILINE_COMMENT;
                            default:
                                assert false :
                                        "bad state "
                                        + category
                                        + " "
                                        + nextChar;
                                return null;
                            } // end switch}
                        }
                    },
        /**
         * Someone has forwarded us an EOL. We deal with it and any subsequent EOLs. When we hit something
         * interesting we
         * let the default next deal with it.
         */
        AT_END_OF_LINE
                    {
                    void consume( char c )
                        {
                        }

                    void leaving()
                        {
                        buildLine();
                        }

                    MASMState next( MASMCharCategory category, char nextChar, boolean first )
                        {
                        //  AT_END_OF_LINE
                        switch ( category )
                            {
                            case EOL:
                                how = HowToProcess.DISCARD;
                                return AT_FIRST_COLUMN;
                            case LT:
                            case GT:
                            case PLAIN:
                            case QUOTE:
                            case SEMICOLON:
                            case SPACE:
                            case TICK:
                                how = HowToProcess.FORWARD;
                                return AT_FIRST_COLUMN;
                            default:
                                assert false :
                                        "bad state "
                                        + category
                                        + " "
                                        + nextChar;
                                return null;
                            } // end switch
                        }
                    };

    /**
     * true if want extra debugging checks and output
     */
    private static final boolean DEBUGGING = false;
    // //////////////////////////////////////////////////////////////////
    // common to all enum constants

    /**
     * configure 0-based column for the third operand column
     */
    private static final int COL_FOR_OPERAND = 25;

    /**
     * configure 0-based column for the second operator column
     */
    private static final int COL_FOR_OPERATOR = 15;

    /**
     * configure 0-based column for the fourth commend column
     */
    private static final int COL_FOR_TAIL_COMMENT = 40;

    /**
     * DOCTYPE for generated HTML
     */
    private static final String DOCTYPE = "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\" " +
                                          "\"http://www.w3.org/TR/html4/loose.dtd\">";

    /**
     * CrLf or lf used for line endings
     */
    private static final String lineSeparator = System.getProperty( "line.separator" );

    /**
     * want a mono font so code will align.
     */
    private static final String stylesheet = "<style type=\"text/css\">\n"
                                             + "body{font-family:Consolas,\"Bitstream Vera Sans Mono\"," +
                                             "\"Lucida Console\",monospace;}\n"
                                             + "span.boundary{color:#000000;font-weight:bold;}\n"
                                             + "span.label{color:#ce0f14;}\n"
                                             + "span.line{color:#5c9548;}\n"
                                             + "span.multi{color:#415fbf;}\n"
                                             + "span.operand{color:#8b7355;}\n"
                                             + "span.operator{color:#000000;}\n"
                                             + "span.tail{color:#5c9548;font-style:italic;}\n"
                                             + "span.depth{color:#808080;background-color:white;}\n"
                                             + "</style>\n";

    /**
     * used to split up the multiline comment
     */
    private static final Pattern SPLIT_ON_NEWLINE = Pattern.compile( "\n" );

    /**
     * accumulates first col
     */
    private static final StringBuilder accumulatedLabel =
            new StringBuilder( 20 );

    /**
     * accumulates comment starting in column 1
     */
    private static final StringBuilder accumulatedLineComment =
            new StringBuilder( 150 );

    /**
     * accumulates a multiline comment, including embedded NLs.
     */
    private static final StringBuilder accumulatedMultilineComment =
            new StringBuilder( 1024 );

    /* accumulates third col
     */
    private static final StringBuilder accumulatedOperand =
            new StringBuilder( 80 );

    /**
     * accumulates second col
     */
    private static final StringBuilder accumulatedOperator =
            new StringBuilder( 20 );

    /**
     * accumulates fourth col
     */
    private static final StringBuilder accumulatedTailComment =
            new StringBuilder( 10 );

    /**
     * what the user wants us to focus on balancing.
     */
    private static Balancer balancer;

    /**
     * how far we are through parsing the program
     */
    private static int charIndex;

    private static int deepestNestingDepthForIf;

    private static int deepestNestingDepthForMacro;

    private static int deepestNestingDepthForProc;

    private static int deepestNestingDepthForSegment;

    /**
     * how the next next character will be treated, usually consumed and stuffed in a buffer, or forwarded to the next
     * state to deal with.
     */
    private static HowToProcess how;

    private static int nestingDepthForIf;

    private static int nestingDepthForMacro;

    private static int nestingDepthForProc;

    private static int nestingDepthForSegment;

    /**
     * used to track blank lines. More than 1 is suppressed
     */
    private static boolean prevLineEmpty = true;

    /**
     * the program or program fragment we are parsing
     */
    private static String program;

    /**
     * used to accumulate one whole indented program in memory
     */
    private static StringBuilder sb;

    /**
     * the length of the program fragment we are parsing.
     */
    private static int size;

    /**
     * spanning chars to show groups
     */
    private static String spanner;
    // methods

    /**
     * build and append one line of HTML, or one multiline comment
     */
    private static void buildLine()
        {
        // Three basic pieces, multiline, line comment, tuple:  label, operator, operand, tail comment..
        final String multiLineComment = accumulatedMultilineComment.toString();
        if ( multiLineComment.length() > 0 )
            {
            buildMultiCommentLine( multiLineComment );
            }
        // might also be a line comment or tuple on the line following the multiline comment.
        final String lineComment = accumulatedLineComment.toString().trim();
        if ( lineComment.length() > 0 )
            {
            buildLineCommentLine( lineComment );
            }
        else
            {
            buildTupleLine();
            }
        }// /method

    /**
     * build a ; line comment
     *
     * @param lineComment text of line comment including ;
     */
    private static void buildLineCommentLine( final String lineComment )
        {
        spanner = buildSpanner();
        if ( sb != null )
            {
            sb.append( mark( spanner, "depth" ) );
            sb.append( mark( lineComment, "line" ) );
            sb.append( lineSeparator );
            }
        accumulatedLineComment.setLength( 0 );
        assert accumulatedLabel.length()
               + accumulatedOperator.length()
               + accumulatedOperand.length()
               == 0 : "bug: line comment with label/operator/operand.";
        }// /method

    /**
     * build multi-line comment line that ends with special char
     *
     * @param multiLineComment the whole multiline commend with embedded \n
     */
    private static void buildMultiCommentLine( final String multiLineComment )
        {
        spanner = buildSpanner();
        if ( sb != null )
            {
            // multiline comment has embedded \n,s, except very end.
            final String[] mls = SPLIT_ON_NEWLINE.split( multiLineComment );
            for ( String ml : mls )
                {
                sb.append( mark( spanner, "spanner" ) );
                sb.append( mark( ml, "multi" ) );
                sb.append( lineSeparator ); // inserts extra separator at end.
                }
            }
        accumulatedMultilineComment.setLength( 0 );
        }// /method

    /**
     * Build the left margin spanner chars to show span of various markers
     *
     * @param nestingDepth    current nesting depth
     * @param maxNestingDepth deepest depth in the whole program
     *
     * @return Some line drawing chars.
     */
    private static String buildOneSpanner( final int nestingDepth, final int maxNestingDepth )
        {
        final StringBuilder sb = new StringBuilder( maxNestingDepth );
        sb.append( ST.rep( '|', Math.max( 0, nestingDepth ) ) );
        sb.append( ST.spaces( Math.max( 0, maxNestingDepth - nestingDepth ) ) );
        return sb.toString();
        }// /method

    /**
     * build spanning lines for left margin.
     */
    private static String buildSpanner()
        {
        switch ( balancer )
            {
            case ALL:
                return "";
            case SEGMENT:
                return buildOneSpanner( nestingDepthForSegment, deepestNestingDepthForSegment );
            case MACRO:
                return buildOneSpanner( nestingDepthForMacro, deepestNestingDepthForMacro );
            case PROC:
                return buildOneSpanner( nestingDepthForProc, deepestNestingDepthForProc );
            case IF:
                return buildOneSpanner( nestingDepthForIf, deepestNestingDepthForIf );
            default:
                throw new IllegalArgumentException( "MASMState program bug. unknown Balancer: " + balancer );
            }
        }// /method

    /**
     * build line made of label, operator, operand tail comment
     */
    private static void buildTupleLine()
        {
        final String label = accumulatedLabel.toString().trim();
        accumulatedLabel.setLength( 0 );
        String operator = accumulatedOperator.toString().trim();
        accumulatedOperator.setLength( 0 );
        operator = operator.toLowerCase();
        final String markedOperator = markOperator( operator );  // calcs nesting depths
        final String operand = accumulatedOperand.toString().trim();
        accumulatedOperand.setLength( 0 );
        final String tailComment = accumulatedTailComment.toString();
        accumulatedTailComment.setLength( 0 );
        // calculate how many spaces needed to separate fields;   Calc on fields without <spans
        int col = label.length();
        final int spacesBeforeOperator = Math.max( 1, COL_FOR_OPERATOR - col );
        col += spacesBeforeOperator;
        col += operator.length();
        final int spacesBeforeOperand = Math.max( 1, COL_FOR_OPERAND - col );
        col += spacesBeforeOperand;
        col += operand.length();
        final int spacesBeforeTailComment = Math.max( 1, COL_FOR_TAIL_COMMENT - col );
        // glue line back together with spaces to make things fit is desired cols as best as possible.
        FastCat fb = new FastCat( 8 );
        if ( label.length() > 0 || operator.length() > 0 || operand.length() > 0 || tailComment.length() > 0 )
            {
            fb.append( mark( label, "label" ) );
            if ( operator.length() > 0 || operand.length() > 0 || tailComment.length() > 0 )
                {
                fb.append( ST.spaces( spacesBeforeOperator ) );
                fb.append( markedOperator );
                if ( operand.length() > 0 || tailComment.length() > 0 )
                    {
                    fb.append( ST.spaces( spacesBeforeOperand ) );
                    fb.append( mark( operand, "operand" ) );
                    if ( tailComment.length() > 0 )
                        {
                        fb.append( ST.spaces( spacesBeforeTailComment ) );
                        fb.append( mark( tailComment, "tail" ) );
                        }
                    }
                }
            }
        final String theLine = fb.toString();
        if ( theLine.length() == 0 )
            {
            // suppress multiple blank lines
            if ( !prevLineEmpty )
                {
                if ( sb != null )
                    {
                    spanner = buildSpanner();
                    sb.append( mark( spanner, "spanner" ) );
                    sb.append( lineSeparator );
                    }
                prevLineEmpty = true;
                }
            }
        else
            {
            if ( sb != null )
                {
                // spanner already computed
                sb.append( mark( spanner, "spanner" ) );
                sb.append( theLine );
                sb.append( lineSeparator );
                }
            prevLineEmpty = false;
            }
        }// /method

    /**
     * pick a colour that suggests the nesting depth of four different
     * bracketers.
     *
     * @return rgb colour
     */
    private static int calcNestingBackgroundColor()
        {
        final int rgb;
        switch ( balancer )
            {
            case ALL:
                if ( nestingDepthForSegment < 0 || nestingDepthForMacro < 0 || nestingDepthForProc < 0 ||
                     nestingDepthForIf < 0 )
                    {
                    rgb = 0xffff00; /* bright yellow */
                    }
                else
                    {
                    int depths = 100 / Math.max( 1, deepestNestingDepthForSegment ) * nestingDepthForSegment;
                    int depthi = 40 / Math.max( 1, deepestNestingDepthForIf ) * nestingDepthForIf;
                    int depthm = 100 / Math.max( 1, deepestNestingDepthForMacro ) * nestingDepthForMacro;
                    int depthp = 50 / Math.max( 1, deepestNestingDepthForProc ) * nestingDepthForProc;
                    // average of individual balancers.
                    final int r = ( ( 255 ) + ( 255 - depthm ) + ( 255 - depthp ) + ( 255 - depthi ) ) / 4;  // no
                    // depths depth1 in all
                    final int g = ( ( 255 - depths ) + ( 255 ) + ( 255 - depthp ) + ( 255 - depthi ) ) / 4;  // no
                    // depthm
                    final int b = ( ( 255 - depths ) + ( 255 - depthm ) + ( 255 ) + ( 255 - depthi ) ) / 4;  // no
                    // depthp
                    rgb = r << 16 | g << 8 | b;
                    }
                break;
            case SEGMENT:
                if ( nestingDepthForSegment < 0 )
                    {
                    rgb = 0xffff00; /* bright yellow */
                    }
                else
                    {
                    int depth = 100 / Math.max( 1, deepestNestingDepthForSegment ) * nestingDepthForSegment;
                    final int r = 255; // deeper red as nest.
                    final int g = 255 - depth;
                    final int b = 255 - depth;
                    rgb = r << 16 | g << 8 | b;
                    }
                break;
            case MACRO:
                if ( nestingDepthForMacro < 0 )
                    {
                    rgb = 0xffff00; /* bright yellow */
                    }
                else
                    {
                    int depth = 100 / Math.max( 1, deepestNestingDepthForMacro ) * nestingDepthForMacro;
                    final int r = 255 - depth;
                    final int g = 255; // deeper green as nest
                    final int b = 255 - depth;
                    rgb = r << 16 | g << 8 | b;
                    }
                break;
            case PROC:
                if ( nestingDepthForProc < 0 )
                    {
                    rgb = 0xffff00; /* bright yellow */
                    }
                else
                    {
                    int depth = 50 / Math.max( 1, deepestNestingDepthForProc ) * nestingDepthForProc;
                    int r = 255 - depth;
                    int g = 255 - depth;
                    int b = 255;  // deeper blue with nest.
                    rgb = r << 16 | g << 8 | b;
                    }
                break;
            case IF:
                if ( nestingDepthForIf < 0 )
                    {
                    rgb = 0xffff00; /* bright yellow */
                    }
                else
                    {
                    int depth = 40 / Math.max( 1, deepestNestingDepthForIf ) * nestingDepthForIf;
                    final int r = 255 - depth;
                    final int g = r;  // grey
                    final int b = r;
                    rgb = r << 16 | g << 8 | b;
                    }
                break;
            default:
                throw new IllegalArgumentException( "program bug: unknown Balancer " + balancer );
            }
        return rgb;
        }// /method

    /**
     * parse text to find maximum nesting depths, used to scale colours and align || spanners down the left side.
     *
     * @param program
     */
    private static void calcNestingDepths( final String program )
        {
        // do a dry run with sb==null so we don't build any strings.
        reset();
        // where we were
        MASMState oldState = AT_FIRST_COLUMN;
        // where we are
        MASMState state = AT_FIRST_COLUMN;
        // were we will be next
        MASMState newState;
        // how is global however, so next can return both a state and
        // how.
        how = null;
        // Note, NO int charIndex !! Don't "repair that".
        // charIndex is a static variable globally known so "isComing" can
        // use
        // it.
        for ( charIndex = 0; charIndex < size; charIndex++ )
            {
            // next char to process
            char theChar = program.charAt( charIndex );
            // decide which general category the char falls in
            final MASMCharCategory category = MASMCharCategory.categorise( theChar );
            if ( category != MASMCharCategory.IGNORE )
                {
                /*
                 * keep going till some state consumes/discards the character.
                 * Allow up to three forwarding attempts to deal with the
                 * character. Usually we should succeed on the first or second
                 * attempt. We always make at least on trip through
                 */
                attempts:
                for ( int times = 0; times < 3; times++ )
                    {
                    // first is deliberately local
                    final boolean first = state != oldState;
                    /*
                     * crank the state machine one cycle, State should modify
                     * how in addition to returning the new state. A little ugly
                     * but simplest way to return a pair of values: state and
                     * how.
                     */
                    how = null;
                    /*
                     * setting to null ensures not setting it will be caught.
                     */
                    /*
                     * This is the guts of the finite state automaton decide the
                     * next state
                     */
                    // vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
                    newState = state.next( category, theChar, first );
                    // ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
                    if ( DEBUGGING )
                        {
                        dumpState( theChar,
                                category,
                                first,
                                oldState,
                                state,
                                newState,
                                how );
                        }
                    // kick over to the next generation,
                    // we are now in the newState. We make the
                    // transition
                    // here.
                    oldState = state;
                    state = newState;
                    if ( how == null )
                        {
                        throw new NullPointerException( "MASMState bug: how not set. OldState=" + oldState + " " +
                                                        "newState=" + state + " next() must not be private." );
                        }
                    switch ( how )
                        {
                        case CONSUME:
                            oldState.consume( theChar );
                            if ( state != oldState )
                                {
                                oldState.leaving();
                                }
                            // and on to the next char
                            break attempts;
                        case DISCARD:
                            if ( state != oldState )
                                {
                                oldState.leaving();
                                }
                            // and on to the next char
                            break attempts;
                        default:
                            // should never get here
                            assert false :
                                    "MASMState failed to set how variable.";
                            break attempts;
                        case FORWARD:
                            assert state
                                   != oldState :
                                    "MASMState attempted to forward a char to the same state.";
                            oldState.leaving();
                            /*
                             * we give that character another try with the new
                             * state
                             */
                        } // end switch
                    /*
                     * we will only ever get here if we are forwarding. The
                     * other case leave the loop early.
                     */
                    }
                // end attempts loop
                // we fall out the bottom and land here no matter what
                assert how == HowToProcess.CONSUME || how == HowToProcess
                        .DISCARD :
                        "MASMState failed to consume char in three state forwarding attempts.";
                } // end if ignore
            } // end for each character
        }// /method

    /**
     * debugging dump system state
     *
     * @param theChar  char we are processing
     * @param category category of the char
     * @param first    did we just enter this state
     * @param oldState old state
     * @param state    current state
     * @param newState next state
     * @param how      do we consume, forward or discard this character.
     */
    private static void dumpState( char theChar,
                                   MASMCharCategory category,
                                   boolean first,
                                   MASMState oldState,
                                   MASMState state,
                                   MASMState newState,
                                   HowToProcess how )
        {
        if ( how == HowToProcess.CONSUME )
            {
            /* use slightly abbreviated form, black */
            out.println( theChar
                         + " "
                         + category
                         + " "
                         + first
                         + " o:"
                         + oldState
                         + " s:"
                         + state
                         + " n:"
                         + newState );
            }
        else
            {/* in red */
            err.println( theChar
                         + " "
                         + category
                         + " "
                         + first
                         + " o:"
                         + oldState
                         + " s:"
                         + state
                         + " n:"
                         + newState
                         + " "
                         + how );
            }
        }// /method

    /**
     * Is a given string isComing up in the stream starting with this character. Compares ignoring case.
     *
     * @param expected string to TEST if isComing in the stream
     *
     * @return true if this string isComing up, case-insensitive
     */
    private static boolean isComing( String expected )
        {
        // check if there are enough characters left for a match.
        return charIndex + expected.length() < size && program.substring(
                charIndex,
                charIndex + expected.length() )
                .equalsIgnoreCase( expected );
        }// /method

    /**
     * add css class to a String
     *
     * @param s        string to decorate
     * @param cssClass class to apply
     *
     * @return HTML to display that string.
     */
    private static String mark( String s, String cssClass )
        {
        if ( s.length() == 0 )
            {
            return "";
            }
        assert cssClass != null && cssClass.length() > 0 : "empty css class";
        final FastCat sb = new FastCat( 5 );
        sb.append( "<span class=\"" );
        sb.append( cssClass );
        sb.append( "\">" );
        sb.append( EntifyStrings.entifyHTML( s ) );
        sb.append( "</span>" );
        return sb.toString();
        }// /method

    /**
     * enclose the operator is HTML span to select its colour based on nesting depth.
     * Calculates nesting depth and spanner
     *
     * @param operator text from second column
     *
     * @return operator enclosed in <span
     */
    private static String markOperator( String operator )
        {
        if ( operator.length() == 0 )
            {
            return "";
            }
        boolean boundary = false;
        if ( operator.equals( "segment" ) || operator.equals( "struc" ) )  // both balance with ends
            {
            nestingDepthForSegment++;
            if ( nestingDepthForSegment > deepestNestingDepthForSegment )
                {
                deepestNestingDepthForSegment = nestingDepthForSegment;
                }
            boundary = true;
            }
        else if ( operator.equals( "macro" ) )
            {
            nestingDepthForMacro++;
            if ( nestingDepthForMacro > deepestNestingDepthForMacro )
                {
                deepestNestingDepthForMacro = nestingDepthForMacro;
                }
            boundary = true;
            }
        else if ( operator.equals( "proc" ) )
            {
            nestingDepthForProc++;
            if ( nestingDepthForProc > deepestNestingDepthForProc )
                {
                deepestNestingDepthForProc = nestingDepthForProc;
                }
            boundary = true;
            }
        else if ( operator.equals( "if" )
                  || operator.equals( "if1" )
                  || operator.equals( "if2" )
                  || operator.equals( "ifb" )
                  || operator.equals( "ifdef" )
                  || operator.equals( "ifdif" )
                  || operator.equals( "ifdifi" )
                  || operator.equals( "ife" )
                  || operator.equals( "ifidn" )
                  || operator.equals( "ifidni" )
                  || operator.equals( "ifnb" )
                  || operator.equals( "ifndef" )
                )
            {
            nestingDepthForIf++;
            if ( nestingDepthForIf > deepestNestingDepthForIf )
                {
                deepestNestingDepthForIf = nestingDepthForIf;
                }
            boundary = true;
            }
        else if ( operator.equals( "else" ) )
            {
            if ( nestingDepthForIf == 0 )
                {
                // if else by itself, is an error.
                nestingDepthForIf = -1;
                }
            boundary = true;
            }
        // We must do various calc here before we start lowering the nesting depths
        final int rgb = calcNestingBackgroundColor();
        spanner = buildSpanner();
        if ( operator.equals( "ends" ) )
            {
            nestingDepthForSegment--;
            boundary = true;
            }
        else if ( operator.equals( "endm" ) )
            {
            nestingDepthForMacro--;
            boundary = true;
            }
        else if ( operator.equals( "endp" ) )
            {
            nestingDepthForProc--;
            boundary = true;
            }
        else if ( operator.equals( "endif" ) )
            {
            nestingDepthForIf--;
            boundary = true;
            }
        else if ( operator.equals( "else" ) )
            {
            if ( nestingDepthForIf == -1 )
                {
                // if else by itself, get back the way it was as if no else there.
                nestingDepthForIf = 0;
                }
            }
        // don't use prevrgb logic since there have been intervening text since the last markOperator
        final FastCat sb = new FastCat( 7 );
        sb.append( "<span class=\"" );
        sb.append( boundary ? "boundary" : "operator" );
        sb.append( "\" style=\"background-color:#" );
        sb.append( ST.toLZHexString( rgb, 6 ) );
        sb.append( ";\">" );
        sb.append( EntifyStrings.entifyHTML( operator ) );
        sb.append( "</span>" );
        return sb.toString();
        }// /method

    /**
     * parse program colourising it.
     *
     * @param program text of the masm program
     *
     * @return colorised HTML text
     */
    private static String parseProgram( final String program )
        {
        reset();
        // where we were
        MASMState oldState = AT_FIRST_COLUMN;
        // where we are
        MASMState state = AT_FIRST_COLUMN;
        // were we will be next
        MASMState newState;
        // how is global however, so next can return both a state and
        // how.
        how = null;
        sb = new StringBuilder( size * 3 + 600 );
        sb.append( DOCTYPE );
        sb.append( "<html lang=\"en-CA\"><head>" );
        sb.append( stylesheet );   // in theory this should not work. Only HTML 3.2 guaranteed to be supported.
        sb.append( "</head><body><pre>" );
        // Note, NO int charIndex !! Don't "repair that".
        // charIndex is a static variable globally known so "isComing" can
        // use
        // it.
        for ( charIndex = 0; charIndex < size; charIndex++ )
            {
            // next char to process
            char theChar = program.charAt( charIndex );
            // decide which general category the char falls in
            final MASMCharCategory category = MASMCharCategory.categorise( theChar );
            if ( category != MASMCharCategory.IGNORE )
                {
                /*
                 * keep going till some state consumes/discards the character.
                 * Allow up to three forwarding attempts to deal with the
                 * character. Usually we should succeed on the first or second
                 * attempt. We always make at least on trip through
                 */
                attempts:
                for ( int times = 0; times < 3; times++ )
                    {
                    // first is deliberately local
                    final boolean first = state != oldState;
                    /*
                     * crank the state machine one cycle, State should modify
                     * how in addition to returning the new state. A little ugly
                     * but simplest way to return a pair of values: state and
                     * how.
                     */
                    how = null;
                    /*
                     * setting to null ensures not setting it will be caught.
                     */
                    /*
                     * This is the guts of the finite state automaton decide the
                     * next state
                     */
                    // vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
                    newState = state.next( category, theChar, first );
                    // ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
                    if ( DEBUGGING )
                        {
                        dumpState( theChar,
                                category,
                                first,
                                oldState,
                                state,
                                newState,
                                how );
                        }
                    // kick over to the next generation,
                    // we are now in the newState. We make the
                    // transition
                    // here.
                    oldState = state;
                    state = newState;
                    if ( how == null )
                        {
                        throw new NullPointerException( "MASMState bug: how not set. OldState=" + oldState + " " +
                                                        "newState=" + state + " next() must not be private." );
                        }
                    switch ( how )
                        {
                        case CONSUME:
                            oldState.consume( theChar );
                            if ( state != oldState )
                                {
                                oldState.leaving();
                                }
                            // and on to the next char
                            break attempts;
                        case DISCARD:
                            if ( state != oldState )
                                {
                                oldState.leaving();
                                }
                            // and on to the next char
                            break attempts;
                        default:
                            // should never get here
                            assert false :
                                    "MASMState failed to set how variable.";
                            break attempts;
                        case FORWARD:
                            assert state
                                   != oldState :
                                    "MASMState attempted to forward a char to the same state.";
                            oldState.leaving();
                            /*
                             * we give that character another try with the new
                             * state
                             */
                        } // end switch
                    /*
                     * we will only ever get here if we are forwarding. The
                     * other case leave the loop early.
                     */
                    }
                // end attempts loop
                // we fall out the bottom and land here no matter what
                assert how == HowToProcess.CONSUME || how == HowToProcess
                        .DISCARD :
                        "MASMState failed to consume char in three state forwarding attempts.";
                } // end if ignore
            } // end for each character
        if ( state != AT_FIRST_COLUMN )
            {
            // force a final wrapup -- e.g.label without terminator
            buildLine();
            }
        sb.append( "</pre></body></html>" );
        final String result = sb.toString();
        sb = null;
        return result;
        }// /method

    /**
     * clear out the state machine ready to parse a new program
     */
    private static void reset()
        {
        nestingDepthForIf = 0;
        nestingDepthForMacro = 0;
        nestingDepthForProc = 0;
        nestingDepthForSegment = 0;
        accumulatedLabel.setLength( 0 );
        accumulatedOperator.setLength( 0 );
        accumulatedOperand.setLength( 0 );
        accumulatedTailComment.setLength( 0 );
        accumulatedLineComment.setLength( 0 );
        accumulatedMultilineComment.setLength( 0 );
        how = null;
        charIndex = 0;
        spanner = "";
        sb = null;
        // keeping certain variables local or global is crucial.
        // Don't mess with them without thinking carefully and
        // changing the docs:
        // local: category, first, oldState, state, theChar
        // global: charIndex, how
        // we make these local to discourage accidental snooping or
        // accidentally picking up the static version instead of the parm.
        }// /method

    /**
     * clear out the state machine ready to parse a new program
     */
    private static void resetDeepestNestingDepths()
        {
        deepestNestingDepthForIf = 0;
        deepestNestingDepthForMacro = 0;
        deepestNestingDepthForProc = 0;
        deepestNestingDepthForSegment = 0;
        }// /method

    /**
     * Default Consume one character. It has been predecided that you can and will consume it.
     *
     * @param c char to consume
     */
    abstract void consume( char c );// /method

    /**
     * default what to do on leaving state, after last char is consumed.
     */
    abstract void leaving();// /method

    /**
     * default next method determines the next state based on current state, and next char
     *
     * @param category class of next character
     * @param nextChar next character to process
     * @param first    true     if we are just entering this state.
     *
     * @return next JavaState
     * D o   n o t   m a k e   p r i v a t e ! ! !
     */
    @SuppressWarnings( { "WeakerAccess" } )
    abstract MASMState next( MASMCharCategory category, char nextChar, boolean first );// /method

    /**
     * Parse program and leave a list of Tokens in tokens ArrayList.
     *
     * @param program  the text we are going parse and eventually render. Its loop calls next on oldstate, then
     *                 optionally consume on the old state, State whose next made the consume decision consumes it own
     *                 char. then optionally leaving on the old state
     * @param balancer which pairs are important to balance.
     *
     * @return the aligned program
     */
    @SuppressWarnings( { "UnusedAssignment" } )
    public static String colourise( String program, Balancer balancer )
        {
        MASMState.program = program;
        size = program.length();
        MASMState.balancer = balancer;
        resetDeepestNestingDepths();
        calcNestingDepths( program );
        /*
         * keeping certain variables local or global is crucial. Don't mess with
         * them without thinking carefully and changing the docs: LOCAL:
         * category, first, oldState, state, theChar. Global: charIndex, how we
         * make these local to discourage accidental snooping or accidentally
         * picking up the static version instead of the parm.
         */
        return parseProgram( program );
        }// /method
    // /methods
    }
